# Licensed under the MIT License.
# i18nco Copyright (C) 2022 numlinka.

DEFAULT_ENCODING = "utf-8"


# writing system | 文字系统
LATIN_ALPHABET = "Latin Alphabet"
CHINESE_CHARACTERS = "Chinese Characters"
CYRILLIC_ALPHABET = "Cyrillic Alphabet"
ARABIC_ALPHABET = "Arabic Alphabet"
DEVANAGARI_ALPHABET = "Devanagari Alphabet"
GREEK_ALPHABET = "Greek Alphabet"
JAPANESE_KANA = "Japanese Kana"
HANGUL = "Hangul"
THAI_ALPHABET = "Thai Alphabet"
HEBREW_ALPHABET = "Hebrew Alphabet"
KHMER_ALPHABET = "Khmer Alphabet"
ETHIOPIC_OR_GE_EZ_SCRIPT = "Ethiopic or Ge'ez Script"
GEORGIAN_ALPHABET = "Georgian Alphabet"
TAMIL_SCRIPT = "Tamil Script"


# Latin Alphabet | 拉丁文
en_US, en_US_lang = "en_US", "English (United States)"
en_GB, en_GB_lang = "en_GB", "English (United Kingdom)"
es_ES, es_ES_lang = "es_ES", "Español (España)"
fr_FR, fr_FR_lang = "fr_FR", "Français (France)"
de_DE, de_DE_lang = "de_DE", "Deutsch (Deutschland)"
it_IT, it_IT_lang = "it_IT", "Italiano (Italia)"
pt_PT, pt_PT_lang = "pt_PT", "Português (Portugal)"
pt_BR, pt_BR_lang = "pt_BR", "Português (Brasil)"
en_SG, en_SG_lang = "en_SG", "English (Singapore)"
ms_SG, ms_SG_lang = "ms_SG", "Bahasa Melayu (Singapura)"
en_MY, en_MY_lang = "en_MY", "English (Malaysia)"
ms_MY, ms_MY_lang = "ms_MY", "Bahasa Melayu (Malaysia)"

# Chinese Characters | 汉字
zh_CN, zh_CN_lang = "zh_CN", "简体中文 (中国)"
zh_HK, zh_HK_lang = "zh_HK", "繁體中文 (中國香港)"
zh_MO, zh_MO_lang = "zh_MO", "繁體中文 (中國澳門)"
zh_TW, zh_TW_lang = "zh_TW", "繁體中文 (中國台灣)"
zh_SG, zh_SG_lang = "zh_SG", "中文 (新加坡)"
zh_MY, zh_MY_lang = "zh_MY", "中文 (马来西亚)"

# Cyrillic Alphabet | 希利尔文
ru_RU, ru_RU_lang = "ru_RU", "Русский (Россия)"
uk_UA, uk_UA_lang = "uk_UA", "Українська (Україна)"
bg_BG, bg_BG_lang = "bg_BG", "Български (България)"
sr_RS, sr_RS_lang = "sr_RS", "Српски (Србија)"

# Arabic Alphabet | 阿拉伯文
ar_SA, ar_SA_lang = "ar_SA", "العربية (المملكة العربية السعودية)"
ar_EG, ar_EG_lang = "ar_EG", "العربية (مصر)"
fa_IR, fa_IR_lang = "fa_IR", "فارسی (ایران)"
ur_PK, ur_PK_lang = "ur_PK", "اُردُو (پاکستان)"

# Devanagari Alphabet | 梵文
hi_IN, hi_IN_lang = "hi_IN", "हिंदी (भारत)"
ne_NP, ne_NP_lang = "ne_NP", "नेपाली (नेपाल)"

# Greek Alphabet | 希腊文
el_GR, el_GR_lang = "el_GR", "Ελληνικά (Ελλάδα)"
el_CY, el_CY_lang = "el_CY", "Ελληνικά (Κύπρος)"

# Japanese Kana | 日文假名
ja_JP, ja_JP_lang = "ja_JP", "日本語"

# Hangul | 韩文
ko_KR, ko_KR_lang = "ko_KR", "한국어 (대한민국)"
ko_KP, ko_KP_lang = "ko_KP", "조선말 (조선 민주주의 인민 공화국)"

# Thai Alphabet | 泰文
th_TH, th_TH_lang = "th_TH", "ไทย (ประเทศไทย)"

# Hebrew Alphabet | 希伯来文
he_IL, he_IL_lang = "he_IL", "עברית (ישראל)"
yi_001, yi_001_lang = "yi_001", "ייִדיש (יידיש)"

# Khmer Alphabet | 柬埔寨文
km_KH, km_KH_lang = "km_KH", "ភាសាខ្មែរ (កម្ពុជា)"

# Ethiopic or Ge'ez Script | 埃塞俄比亚文
am_ET, am_ET_lang = "am_ET", "አማርኛ (ኢትዮጵያ)"
ti_ER, ti_ER_lang = "ti_ER", "ትግርኛ (ኤርትራ)"
et_EE, et_EE_lang = "et_EE", "አማርኛ (ኢትዮጵያ)"

# Georgian Alphabet | 格鲁吉亚文
ka_GE, ka_GE_lang = "ka_GE", "ქართული (საქართველო)"

# Tamil Script | 泰米尔文
ta_IN, ta_IN_lang = "ta_IN", "தமிழ் (இந்தியா)"
ta_LK, ta_LK_lang = "ta_LK", "தமிழ் (இலங்கை)"
ta_MY, ta_MY_lang = "ta_MY", "தமிழ் (மலேசியா)"
ta_SG, ta_SG_lang = "ta_SG", "தமிழ் (சிங்கப்பூர்)"


LATIN_ALPHABET_TABLE = {
    en_US: en_US_lang,
    en_GB: en_GB_lang,
    es_ES: es_ES_lang,
    fr_FR: fr_FR_lang,
    de_DE: de_DE_lang,
    it_IT: it_IT_lang,
    pt_PT: pt_PT_lang,
    pt_BR: pt_BR_lang,
    en_SG: en_SG_lang,
    ms_SG: ms_SG_lang,
    en_MY: en_MY_lang,
    ms_MY: ms_MY_lang,
}

CHINESE_CHARACTERS_TABLE = {
    zh_CN: zh_CN_lang,
    zh_HK: zh_HK_lang,
    zh_MO: zh_MO_lang,
    zh_TW: zh_TW_lang,
    zh_SG: zh_SG_lang,
    zh_MY: zh_MY_lang,
    ja_JP: ja_JP_lang,
    ko_KR: ko_KR_lang,
    ko_KP: ko_KP_lang,
}

CYRILLIC_ALPHABET_TABLE = {
    ru_RU: ru_RU_lang,
    uk_UA: uk_UA_lang,
    bg_BG: bg_BG_lang,
    sr_RS: sr_RS_lang,
}

ARABIC_ALPHABET_TABLE = {
    ar_SA: ar_SA_lang,
    ar_EG: ar_EG_lang,
    fa_IR: fa_IR_lang,
    ur_PK: ur_PK_lang,
}

DEVANAGARI_ALPHABET_TABLE = {
    hi_IN: hi_IN_lang,
    ne_NP: ne_NP_lang,
}

GREEK_ALPHABET_TABLE = {
    el_GR: el_GR_lang,
    el_CY: el_CY_lang,
}

JAPANESE_KANA_TABLE = {
    ja_JP: ja_JP_lang,
}

HANGUL_TABLE = {
    ko_KR: ko_KR_lang,
    ko_KP: ko_KP_lang,
}

THAI_ALPHABET_TABLE = {
    th_TH: th_TH_lang,
}

HEBREW_ALPHABET_TABLE = {
    he_IL: he_IL_lang,
    yi_001: yi_001_lang,
}

KHMER_ALPHABET_TABLE = {
    km_KH: km_KH_lang,
}

ETHIOPIC_OR_GE_EZ_SCRIPT_TABLE = {
    am_ET: am_ET_lang,
    ti_ER: ti_ER_lang,
    et_EE: et_EE_lang,
}

GEORGIAN_ALPHABET_TABLE = {
    ka_GE: ka_GE_lang,
}

TAMIL_SCRIPT_TABLE = {
    ta_IN: ta_IN_lang,
    ta_LK: ta_LK_lang,
    ta_MY: ta_MY_lang,
    ta_SG: ta_SG_lang,
}


LANGUAGE_TABLE = {
    **LATIN_ALPHABET_TABLE,
    **CHINESE_CHARACTERS_TABLE,
    **CYRILLIC_ALPHABET_TABLE,
    **ARABIC_ALPHABET_TABLE,
    **DEVANAGARI_ALPHABET_TABLE,
    **GREEK_ALPHABET_TABLE,
    **JAPANESE_KANA_TABLE,
    **HANGUL_TABLE,
    **THAI_ALPHABET_TABLE,
    **HEBREW_ALPHABET_TABLE,
    **KHMER_ALPHABET_TABLE,
    **ETHIOPIC_OR_GE_EZ_SCRIPT_TABLE,
    **GEORGIAN_ALPHABET_TABLE,
    **TAMIL_SCRIPT_TABLE,
}

WRITING_SYSTEM_TABLE = {
    LATIN_ALPHABET: LATIN_ALPHABET_TABLE,
    CHINESE_CHARACTERS: CHINESE_CHARACTERS_TABLE,
    CYRILLIC_ALPHABET: CYRILLIC_ALPHABET_TABLE,
    ARABIC_ALPHABET: ARABIC_ALPHABET_TABLE,
    DEVANAGARI_ALPHABET: DEVANAGARI_ALPHABET_TABLE,
    GREEK_ALPHABET: GREEK_ALPHABET_TABLE,
    JAPANESE_KANA: JAPANESE_KANA_TABLE,
    HANGUL: HANGUL_TABLE,
    THAI_ALPHABET: THAI_ALPHABET_TABLE,
    HEBREW_ALPHABET: HEBREW_ALPHABET_TABLE,
    KHMER_ALPHABET: KHMER_ALPHABET_TABLE,
    ETHIOPIC_OR_GE_EZ_SCRIPT: ETHIOPIC_OR_GE_EZ_SCRIPT_TABLE,
}

UNITED_NATIONS_LANGUAGE_LIST = ("en", "fr", "es", "zh", "ru")


__all__ = [x for x in globals() if not x.startswith("_")]



# Licensed under the MIT License.
# i18nco Copyright (C) 2022 numlinka.

__all__ = [
    "I18nString",
    "Internationalization"
]

__name__ = "internationalization"
__author__ = "numlinka"
__license__ = "MIT"
__copyright__ = "Copyright (C) 2022 numlinka"
__version_info__ = (1, 5, 0)
__version__ = ".".join(map(str, __version_info__))

# internal
from .internationalization import I18nString, Internationalization
from . import utils



# Licensed under the MIT License.
# i18nco Copyright (C) 2022 numlinka.

__all__ = [
    "decode_escape_sequences",
    "get_locale_code",
    "match_best_locale"
]

# std
import os
import re
import sys
import ctypes
import locale
from typing import List, Optional

# self
from .typeins import LocaleCode
from .constants import en_US, WRITING_SYSTEM_TABLE, UNITED_NATIONS_LANGUAGE_LIST


ESCAPE_SEQUENCE_RE = re.compile("\\\\(u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8}|x[0-9A-Fa-f]{2}|.)")


def decode_escape_sequences(text: str) -> str:
    """
    Decodes escape sequences in a string to their corresponding characters.

    Supports multiple escape sequence formats including:
    - Unicode: \\uXXXX (16-bit), \\UXXXXXXXX (32-bit)
    - Hexadecimal: \\xXX
    - Common escape characters: \\n, \\t, \\b, \\r, \\f, \\a, \\v
    - Special cases: \\\\ (backslash), \\0 (null), \\_ (space)
    - Unrecognized escape sequences are preserved as-is

    Args:
        text (str): Input string containing escape sequences

    Returns:
        out (str): Decoded string with all escape sequences replaced by their corresponding characters
    """
    def replace_escape(match):
        escape = match.group(1)
        if escape.startswith("u") or escape.startswith("U"):
            return chr(int(escape[1:], 16))
        elif escape.startswith("x"):
            return chr(int(escape[1:], 16))
        elif escape == "n":
            return "\n"
        elif escape == "t":
            return "\t"
        elif escape == "b":
            return "\b"
        elif escape == "r":
            return "\r"
        elif escape == "f":
            return "\f"
        elif escape == "a":
            return "\a"
        elif escape == "v":
            return "\v"
        elif escape == "0":
            return "\0"
        elif escape == "\\":
            return "\\"
        elif escape == "_":
            return " "
        else:
            return escape

    return ESCAPE_SEQUENCE_RE.sub(replace_escape, text)


def get_locale_code() -> LocaleCode:
    """
    Get the system locale code in the format "lang_COUNTRY" (e.g., "en_US").
    Defaults to "en_US" if no locale can be determined.

    Returns:
        out (LocaleCode): The system locale code in the format "lang_COUNTRY" (e.g., "en_US").
    """
    locale_info = os.environ.get("LANG", "")
    locale_code = locale_info.split(".")[0] if locale_info else ""
    if locale_code:
        return locale_code

    try:
        locale_code = locale.getdefaultlocale()[0]
        if locale_code:
            return locale_code

    except AttributeError:
        pass

    if sys.platform == "win32":
        try:
            windll = ctypes.windll.kernel32
            language_id = windll.GetUserDefaultUILanguage()
            locale_code = locale.windows_locale.get(language_id, "")
            if locale_code:
                return locale_code

        except Exception:
            pass

    return en_US


def match_best_locale(target: LocaleCode, available: List[LocaleCode]) -> Optional[LocaleCode]:
    """
    Finds the best matching locale from available options for a target locale.

    The matching is performed in the following priority order:
    1. Exact match (e.g., target "en_US" matches available "en_US")
    2. Language code match (e.g., target "en_US" matches available "en_BG")
    3. Writing system match (e.g., target "en_US" matches available "de_DE")

    Args:
        target (LocaleCode): The target locale code to match against (e.g., "en_US")
        available (List[LocaleCode]): List of available locale codes (e.g., ["en_US", "en_GB"])

    Returns:
        out (LocaleCode | None): The best matching locale code from available options, or `None` if no match found.
    """
    if not isinstance(target, LocaleCode):
        raise TypeError(
            f"Expected `target` to be of type `LocaleCode`, but got {type(target)}.")

    if not isinstance(available, (list, tuple)) or not all(isinstance(locale, LocaleCode) for locale in available):
        raise TypeError(
            f"Expected `available` to be a `list` of `LocaleCode`, but got {type(available)}.")

    if target in available:
        return target

    language_code = target.split("_")[0]
    for locale_code in available:
        if locale_code.startswith(language_code):
            return locale_code

    for _, table in WRITING_SYSTEM_TABLE.items():
        if target not in table:
            continue

        for locale_code in table:
            if locale_code in available:
                return locale_code

    for language_code in UNITED_NATIONS_LANGUAGE_LIST:
        for locale_code in available:
            if locale_code.startswith(language_code):
                return locale_code

    return None



# Licensed under the MIT License.
# i18nco Copyright (C) 2022 numlinka.

__all__ = ["Internationalization", "I18nString"]

# std
import os
import csv
import json
import threading
from typing import Tuple, List, Dict, Optional, Union

# internal
from .utils import match_best_locale, decode_escape_sequences
from .typeins import TextKey, LocaleCode
from .constants import DEFAULT_ENCODING, en_US


class I18nString (str):
    """
    A specialized string class for handling internationalized (i18n) text.

    This class extends the built-in `str` class to support dynamic translation lookups
    and attribute-based access to nested translation keys. It is designed to work in
    conjunction with the `Internationalization` class to provide seamless multi-language
    support.
    """

    def _ctrl_relevance(self, visit: "Internationalization", keywords: str = "") -> None:
        """
        Set the `Internationalization` instance and the base translation key.

        This method associates the `I18nString` instance with an `Internationalization`
        object and sets the base translation key. The key can be used for nested
        translation lookups.

        Args:
            visit (Internationalization): The `Internationalization` instance to use for
                                          translation lookups.
            keywords (str, optional): The base translation key or path. Defaults to "".
        """
        self._visit_ = visit
        self._keywords_ = keywords

    def __getattr__(self, __name: str) -> "I18nString":
        """
        Dynamically resolve nested translation keys.

        This method allows dot notation (e.g., "welcome.message") to access nested
        translation keys. If the current `_keywords_` is not empty, it appends the new
        key to the existing path.

        Args:
            __name (str): The attribute name (translation key) to resolve.

        Returns:
            out (I18nString): A new `I18nString` instance representing the resolved translation.
        """
        target = __name if self._keywords_ == "" else f"{self._keywords_}.{__name}"
        return self._visit_.ctrl.translation(target)


    def __getitem__(self, __name: str) -> "I18nString":
        """
        Dynamically resolve nested translation keys.

        This method allows dot notation (e.g., "welcome.message") to access nested
        translation keys. If the current `_keywords_` is not empty, it appends the new
        key to the existing path.

        Args:
            __name (str): The attribute name (translation key) to resolve.

        Returns:
            out (I18nString): A new `I18nString` instance representing the resolved translation.
        """
        target = __name if self._keywords_ == "" else f"{self._keywords_}.{__name}"
        return self._visit_.ctrl.translation(target)


class Internationalization (object):
    def __init__(self) -> None:
        """
        This class provides an interface for managing internationalization (i18n) translations.
        """
        self.ctrl = I18nControl(self)

    def _attribute_overload(self, target: TextKey) -> str:
        if not isinstance(target, TextKey):
            raise TypeError(f"Expected target to be TextKey, but got {type(target)}.")

        try:
            result: Optional[str] = None
            for indexed, __name in enumerate(target.split(".")):
                result = super().__getattribute__(__name) if indexed == 0 else getattr(result, __name)

            if not isinstance(result, str):
                result = target

        except AttributeError as _:
            result = target

        return result

    def __getattribute__(self, __name: str) -> I18nString:
        if __name.startswith("_") or __name == "ctrl":
            return super().__getattribute__(__name)

        return self.ctrl.translation(__name)

    def __getitem__(self, __name: str) -> I18nString:
        return self.ctrl.translation(__name)

    def __call__(self, key: TextKey, locale: Optional[LocaleCode] = None) -> I18nString:
        return self.ctrl.translation(key, locale)



class I18nControl (object):
    def __init__(self, i18n: Internationalization) -> None:
        if not isinstance(i18n, Internationalization):
            raise TypeError(f"Expected `i18n` to be an instance of `Internationalization`, but got {type(i18n)}.")

        self.i18n = i18n
        self._lock = threading.RLock()
        self.__table_translation: Dict[LocaleCode: Dict[TextKey: str]] = {}
        self.__first_locale: LocaleCode = en_US
        self.__second_locale: LocaleCode = en_US

    @property
    def first_locale(self) -> LocaleCode:
        with self._lock:
            return self.__first_locale

    @property
    def second_locale(self) -> LocaleCode:
        with self._lock:
            return self.__second_locale

    @property
    def available_locales(self) -> Tuple[LocaleCode]:
        with self._lock:
            return tuple(self.__table_translation.keys())

    def set_locale(self, first_language: Optional[LocaleCode] = None,
            second_language: Optional[LocaleCode] = None, auto_adjust: bool = True) -> None:
        """
        Set the locale for language preferences.

        You can set two languages: a primary language (`first_language`) and a secondary language (`second_language`).
        The secondary language is used when a valid translation is not available in the primary language.
        By default, the secondary language is automatically adjusted based on the primary language.

        Args:
            first_language (LocaleCode, optional): The primary language to use. If None, the existing primary language
                                                   is retained.
            second_language (LocaleCode, optional): The secondary language to use. If None, the secondary language is
                                                    automatically adjusted based on the primary language.
            auto_adjust (bool): Whether to automatically adjust the secondary language based on the primary language.
                                This parameter has no effect if `second_language` is explicitly set.

        Raises:
            TypeError: If `first_language` or `second_language` is not a `LocaleCode` or None.

        Examples:
            Set only the primary language:
            >>> obj.set_locale(zh_CN, auto_adjust=False)

            Set only the secondary language:
            >>> obj.set_locale(second_language=zh_CN)

            Set both primary and secondary languages:
            >>> obj.set_locale(en_US, zh_CN)

            Set the primary language and automatically adjust the secondary language:
            >>> obj.set_locale(zh_CN)

            Automatically adjust the secondary language based on the existing primary language:
            >>> obj.set_locale()
        """
        if first_language is not None and not isinstance(first_language, LocaleCode):
            raise TypeError(f"Expected `first_language` to be `LocaleCode` or None, but got {type(first_language)}.")

        if second_language is not None and not isinstance(second_language, LocaleCode):
            raise TypeError(f"Expected `second_language` to be `LocaleCode` or None, but got {type(second_language)}.")

        with self._lock:
            if isinstance(first_language, LocaleCode):
                self.__first_locale = first_language

            if isinstance(second_language, LocaleCode):
                self.__second_locale = second_language

            elif second_language is None and auto_adjust:
                available_locales = [x for x in self.available_locales if x != self.__first_locale]
                best_locale = match_best_locale(self.__first_locale, available_locales)

                if best_locale is None:
                    self.__second_locale = en_US
                    return

                self.__second_locale = best_locale


    def set_translation(self, locale: Optional[Union[LocaleCode, List[LocaleCode]]] = None,
                        key: Optional[TextKey] = None, text: Optional[str] = None) -> None:
        """
        Set or remove the translation text for the specified language key.

        This method allows you to set or remove a translation for a specific key and language. If `text` is None,
        the translation for the specified key will be removed. If `key` is None, all translations for the specified
        language will be removed.

        Args:
            locale (LocaleCode, List[LocaleCode], optional): The language for which to set or remove the translation.
                                                         If None, the `first_language` (previously set) will be used.
            key (TextKey, optional): The key of the translation to set or remove. If None, all translations for the
                                 specified language will be removed.
            text (str, optional): The text to set for the translation. If None, the translation for the specified key
                              will be removed.

        Raises:
            TypeError: If the input types are incorrect.

        Examples:
            Set a translation for a specific key and language:
            >>> obj.set_translation(LocaleCode("en_US"), "welcome_message", "Welcome!")

            Set a translation for multiple languages:
            >>> obj.set_translation([LocaleCode("en_US"), LocaleCode("zh_CN")], "welcome_message", "Welcome!")

            Remove a translation for a specific key:
            >>> obj.set_translation(LocaleCode("en_US"), "welcome_message", None)

            Remove all translations for a specific language:
            >>> obj.set_translation(LocaleCode("en_US"), None, None)

            Set a translation using the default `first_language`:
            >>> obj.set_translation(None, "welcome_message", "Welcome!")
        """
        # if locale is not None and not isinstance(locale, LocaleCode):
        #     raise TypeError(f"Expected `locale` to be `LocaleCode` or `None`, but got {type(locale)}.")

        if key is not None and not isinstance(key, TextKey):
            raise TypeError(f"Expected `key` to be `TextKey` or `None`, but got {type(key)}.")

        if text is not None and not isinstance(text, str):
            raise TypeError(f"Expected `text` to be `str` or `None`, but got {type(text)}.")

        # TODO 判断可迭代对象
        # ! 这里使用 isinstance(locale, Iterable) 会出现无限递归错误,
        # ! 无论这个 Iterable 是来自于 collections 还是 typing.
        if isinstance(locale, (list, tuple)):
            for locale_code in locale:
                self.set_translation(locale_code, key, text)
            return

        if locale is not None and not isinstance(locale, LocaleCode):
            raise TypeError(f"Expected `locale` to be `LocaleCode` or `None`, but got {type(locale)}.")

        with self._lock:
            if locale is None:
                locale = self.first_locale

            if key is None:
                if locale in self.__table_translation:
                    del self.__table_translation[locale]
                return

            if text is None:
                if locale in self.__table_translation and key in self.__table_translation[locale]:
                    del self.__table_translation[locale][key]
                return

            if locale not in self.__table_translation:
                self.__table_translation[locale] = {}

            self.__table_translation[locale][key] = text

    def translation(self, key: TextKey, locale: Optional[LocaleCode] = None) -> I18nString:
        """
        Get the translation text for the specified language key.

        This method retrieves the translated text for a given key in the specified language. If the `locale` parameter
        is provided, the translation will be fetched for that specific language.

        If no translation is found for the specified key, an `I18nString` object containing the key itself will
        be returned.

        Args:
            key (TextKey): The key of the translation text to retrieve.
            locale (LocaleCode, optional): The language for which to retrieve the translation.

        Returns:
            out (I18nString): The translated text for the specified key and language. If no translation is found,
                              an `I18nString` object containing the key itself will be returned.

        Raises:
            TypeError: If the input types are incorrect.

        Examples:
            Get a translation for a specific key and language:
            >>> translation = obj.translation("welcome_message", LocaleCode("en_US"))
            >>> print(translation)
            "Welcome!"

            Get a translation without specifying a locale (uses default or fallback language):
            >>> translation = obj.translation("welcome_message")
            >>> print(translation)
            "欢迎！"

            Get a translation for a non-existent key:
            >>> translation = obj.translation("unknown_key")
            >>> print(translation)
            "unknown_key"  # Returns the key itself as an I18nString object.
        """
        if not isinstance(key, TextKey):
            raise TypeError(f"Expected `key` to be `TextKey`, but got {type(key)}.")

        if locale is not None and not isinstance(locale, LocaleCode):
            raise TypeError(f"Expected `locale` to be `LocaleCode` or `None`, but got {type(locale)}.")

        with self._lock:
            target_locale = locale if locale is not None else self.first_locale
            result = self.__table_translation.get(target_locale, {}).get(key, None)

            if result is None and locale is None:
                result = self.__table_translation.get(self.second_locale, {}).get(key, None)

            if result is None:
                result = self.i18n._attribute_overload(key)

            reply = I18nString(result)
            reply._ctrl_relevance(self.i18n, key)
            return reply

    def load_lang_content(self, content: str, locale: Optional[LocaleCode] = None,
                          superiors: Optional[str] = None) -> None:
        """
        Load language content from a string.

        This method parses the provided language file content and loads it into the current object. If `locale` is
        specified, the content will be associated with that locale; otherwise, the default locale will be used.

        The `superiors` parameter allows you to specify a default prefix for all keys in the content.

        Args:
            content (str): The content of the language file to load.
            locale (LocaleCode, optional): The locale to associate with the content. If None, the default locale
                                           will be used.
            superiors (str, optional): A default prefix to add to all keys in the content.

        Raises:
            TypeError: If `locale` is not an instance of `LocaleCode` or `superiors` is not a string.

        Examples:
            Load content for a specific locale:
            >>> content = '''
            ... welcome = Welcome!
            ... '''
            >>> obj.load_lang_content(content, LocaleCode("en_US"))
            >>> obj.translation("welcome")
            "Welcome!"

            Load content with a default prefix:
            >>> content = '''
            ... welcome = Welcome!
            ... '''
            >>> obj.load_lang_content(content, superiors="app")
            >>> obj.translation("app.welcome")
            "Welcome!"

            Load content with a default prefix and override it using `#define`:
            >>> content = '''
            ... #define superiors custom
            ... welcome = Welcome!
            ... '''
            >>> obj.load_lang_content(content, superiors="app")
            >>> obj.translation("custom.welcome")
            "Welcome!"
            """
        if not isinstance(content, str):
            raise TypeError(f"Expected `content` to be `str`, but got {type(content)}.")

        elif not content:
            return

        if locale is None:
            locale = self.i18n.ctrl_get_locale()[0]

        elif not isinstance(locale, LocaleCode):
            raise TypeError(f"Expected `locale` to be `LocaleCode`, but got {type(locale)}.")

        if superiors is None:
            superiors = ""

        elif not isinstance(superiors, str):
            raise TypeError(f"Expected `superiors` to be `str`, but got {type(superiors)}.")

        content_lines = content.splitlines()
        multiline_mode = False
        multiline_line = []

        for line in content_lines:
            line = line.strip()

            if not multiline_mode:
                if line.startswith("#define"):
                    instruction = line.split(" ")
                    if len(instruction) <= 2:
                        continue

                    define = instruction[1]
                    values = [x for x in instruction[2:] if x]

                    if define == "locale":
                        if not values:
                            continue

                        locale = values

                    elif define == "superiors":
                        if not values or values[0] in [".", "/", "#"]:
                            superiors = ""

                        elif isinstance(values, str):
                            superiors = values

                        elif isinstance(values, list):
                            superiors = ".".join(values)

                    continue

                elif line.startswith("#"):
                    continue

                elif line.startswith(";"):
                    continue

                elif line.startswith("//"):
                    continue

                else:
                    lst = line.split("=", 1)
                    if len(lst) != 2:
                        continue

                    key = lst[0].strip()
                    text = lst[1].strip()

            else:
                text = line.strip()

            if text.endswith(" \\") or text == "\\":
                multiline_mode = True
                text = text[:-2]

            else:
                multiline_mode = False

            if len(text) >= 2 and text.startswith("\"") and text.endswith("\""):
                text = text[1:-1]

            if multiline_mode:
                multiline_line.append(text)
                continue

            final_key = f"{superiors}.{key}" if superiors != "" else key

            if multiline_line:
                multiline_line.append(text)
                text = "".join(multiline_line)
                multiline_line = []

            text = decode_escape_sequences(text)
            self.set_translation(locale, final_key, text)

        else:
            if multiline_line:
                final_key = f"{superiors}.{key}" if superiors != "" else key
                text = "".join(multiline_line)
                text = decode_escape_sequences(text)
                self.set_translation(locale, final_key, text)

    def load_lang(self, file_path: str, locale: Optional[LocaleCode] = None, superiors: Optional[str] = None,
                  *, encoding: str = DEFAULT_ENCODING) -> None:
        """
        Load language content from a file.

        This method reads the content of a language file and loads it into the current object. If `locale` is specified,
        the content will be associated with that locale; otherwise, the default locale will be used.

        The `superiors` parameter allows you to specify a default prefix for all keys in the content.

        Args:
            file_path (str): The path to the language file to load.
            locale (LocaleCode, optional): The locale to associate with the content. If None, the default locale
                                        will be used.
            superiors (str, optional): A default prefix to add to all keys in the content.
            encoding (str, optional): The encoding to use when reading the file. Defaults to `DEFAULT_ENCODING`.

        Raises:
            FileNotFoundError: If the specified file does not exist.
            ValueError: If the file content is empty or invalid.
            TypeError: If `locale` is not an instance of `LocaleCode` or `superiors` is not a string.

        Examples:
            Load content for a specific locale:
            >>> obj.load_lang("path/to/lang_file.txt", LocaleCode("en_US"))

            Load content with a default prefix:
            >>> obj.load_lang("path/to/lang_file.txt", superiors="app")

            Load content with a custom encoding:
            >>> obj.load_lang("path/to/lang_file.txt", encoding="utf-16")
        """
        with open(file_path, "r", encoding=encoding) as file_object:
            content = file_object.read()

        self.load_lang_content(content=content, locale=locale, superiors=superiors)

    def load_csv_i18n(self, file_path: str, *, encoding: str = DEFAULT_ENCODING) -> None:
        """
        Load internationalization content from a CSV file.

        This method reads the content of a CSV file and loads it into the current object. The CSV file must have a header
        row with at least three columns: `locale`, `key`, and `value`. Each row represents a translation entry.

        Args:
            file_path (str): The path to the CSV file to load.
            encoding (str, optional): The encoding to use when reading the file. Defaults to `DEFAULT_ENCODING`.

        Raises:
            FileNotFoundError: If the specified file does not exist.
            ValueError: If the file content is empty, invalid, or improperly formatted.
            UnicodeDecodeError: If the file cannot be decoded with the specified encoding.

        Examples:
            Load content from a CSV file:
            >>> obj.load_csv_i18n("path/to/translations.csv")

            Load content from a CSV file with a custom encoding:
            >>> obj.load_csv_i18n("path/to/translations.csv", encoding="utf-16")

            Example CSV file format:
            >>> locale,key,value
            ... en_US,greeting,Hello
            ... zh_CN,greeting,你好
        """
        with open(file_path, "r", encoding=encoding) as file_object:
            reader = csv.DictReader(file_object)

        for row in reader:
            row: dict
            locale = row["locale"]
            key = row["key"]
            value = row["value"]
            value = decode_escape_sequences(value)

            self.set_translation(locale, key, value)

    def load_dict(self, dictionary: Dict[TextKey, str], locale: LocaleCode, superiors: Optional[str] = None) -> None:
        """
        Load translation content from a dictionary.

        This method loads translation key-value pairs from a dictionary into the current object. The translations will be
        associated with the specified locale. If `superiors` is provided, it will be used as a default prefix for all keys.

        Args:
            dictionary (Dict[TextKey, str]): A dictionary containing translation key-value pairs.
            locale (LocaleCode): The locale to associate with the translations.
            superiors (str, optional): A default prefix to add to all keys. If a key is prefixed with `#define`,
                                    it will override this setting.

        Raises:
            ValueError: If the dictionary is empty or contains invalid data.
            TypeError: If `locale` is not an instance of `LocaleCode` or `superiors` is not a string.

        Examples:
            Load translations for a specific locale:
            >>> translations = {"greeting": "Hello", "farewell": "Goodbye"}
            >>> obj.load_dict(translations, LocaleCode("en_US"))

            Load translations with a default prefix:
            >>> translations = {"greeting": "Hello", "farewell": "Goodbye"}
            >>> obj.load_dict(translations, LocaleCode("en_US"), superiors="app")
        """
        if superiors is None:
            superiors = ""

        elif not isinstance(superiors, str):
            raise TypeError(f"Expected `superiors` to be of type `str`, but got {type(superiors)}.")

        for key, value in dictionary.items():
            final_key = f"{superiors}.{key}" if superiors != "" else key

            if isinstance(value, str):
                self.set_translation(locale, final_key, value)

            elif isinstance(value, dict):
                self.load_dict(value, locale, final_key)

    def load_json(self, file_path: str, locale: Optional[LocaleCode] = None, *, encoding: str = DEFAULT_ENCODING) -> None:
        """
        Load translation content from a JSON file.

        This method reads the content of a JSON file and loads it into the current object. If `locale` is specified,
        the content will be associated with that locale; otherwise, the default locale will be used.

        The JSON file should be formatted as a dictionary where keys are translation keys and values are translation texts.

        Args:
            file_path (str): The path to the JSON file to load.
            locale (LocaleCode, optional): The locale to associate with the content. If None, the default locale
                                           will be used.
            encoding (str, optional): The encoding to use when reading the file. Defaults to `DEFAULT_ENCODING`.

        Raises:
            FileNotFoundError: If the specified file does not exist.
            ValueError: If the file content is empty, invalid, or improperly formatted.
            TypeError: If `locale` is not an instance of `LocaleCode`.
            json.JSONDecodeError: If the file content is not valid JSON.

        Examples:
            Load content for a specific locale:
            >>> obj.load_json("path/to/translations.json", LocaleCode('en_US'))

            Load content without specifying a locale (uses default locale):
            >>> obj.load_json("path/to/translations.json")

            Load content with a custom encoding:
            >>> obj.load_json("path/to/translations.json", encoding="utf-16")
        """
        if superiors is None:
            superiors = ""

        elif not isinstance(superiors, str):
            raise TypeError(f"Excepted `superiors` to be of type `str`, but got {type(superiors)}.")

        with open(file_path, "r", encoding=encoding) as file_object:
            content = file_object.read()

        data = json.loads(content)
        self.load_dict(data, locale)

    def load_json_i18n(self, file_path: str, *, encoding: str = DEFAULT_ENCODING) -> None:
        """
        Load internationalization content from a JSON file.

        This method reads the content of a JSON file and loads it into the current object. The JSON file should be formatted
        as a dictionary where keys are locales and values are dictionaries of translation key-value pairs.

        Args:
            file_path (str): The path to the JSON file to load.
            encoding (str, optional): The encoding to use when reading the file. Defaults to `DEFAULT_ENCODING`.

        Raises:
            FileNotFoundError: If the specified file does not exist.
            ValueError: If the file content is empty, invalid, or improperly formatted.
            json.JSONDecodeError: If the file content is not valid JSON.

        Examples:
            Load content from a JSON file:
            >>> obj.load_json_i18n("path/to/translations.json")

            Load content from a JSON file with a custom encoding:
            >>> obj.load_json_i18n("path/to/translations.json", encoding="utf-16")

            Example JSON file format:
            >>> {
            ...     "en_US": {
            ...         "greeting": "Hello",
            ...         "farewell": "Goodbye"
            ...     },
            ...     "zh_CN": {
            ...         "greeting": "你好",
            ...         "farewell": "再见"
            ...     }
            ... }
        """
        with open(file_path, "r", encoding=encoding) as file_object:
            content = file_object.read()

        data = json.loads(content)
        for locale, dictionary in data.items():
            self.load_dict(dictionary, locale)

    def auto_load(self, path: str, *, locale: Optional[LocaleCode] = None) -> None:
        """
        Automatically load all language files from a directory.

        This method scans the specified directory and loads all supported language files (e.g., `.lang`, `.json`, `.csv`).
        The behavior depends on whether the `locale` parameter is specified:

        1. If `locale` is specified:
        - The directory should contain files without locale codes in their names.
        - All files will be associated with the specified `locale`.
        - Example structure:
            ```
            path
            ├── xxx.lang
            ├── xxx.lang
            ├── xxx.json
            └── xxx.csv
            ```

        2. If `locale` is not specified:
        - The directory can be structured in one of the following ways:
            - Flat structure: Files contain locale codes in their names (e.g., `zh_CN.lang`).
            ```
            path
            ├── zh_CN.lang
            ├── en_US.lang
            ├── ru_RU.lang
            ├── zh_CN.json
            └── xxx.csv
            ```
            - Nested structure: Files are organized into subdirectories named by locale codes.
            ```
            path
            ├── zh_CN
            │   ├── xxx.lang
            │   ├── xxx.lang
            │   └── xxx.json
            └── en_US
                ├── xxx.lang
                ├── xxx.lang
                └── xxx.json
            ```
        - The locale will be inferred from the file names or subdirectory names.

        Args:
            path (str): The path to the directory containing language files.
            locale (LocaleCode, optional): The locale to associate with the loaded content. If None, the locale will be
                                        inferred from the file names or subdirectory names.

        Raises:
            FileNotFoundError: If the specified directory does not exist.
            ValueError: If the directory is empty, contains no supported language files, or file/directory names are invalid.
            TypeError: If `locale` is not an instance of `LocaleCode`.

        Examples:
            Load files for a specific locale:
            >>> obj.auto_load("path/to/directory", locale=LocaleCode("en_US"))

            Load files and infer locale from file names (flat structure):
            >>> obj.auto_load("path/to/flat_directory")

            Load files and infer locale from subdirectory names (nested structure):
            >>> obj.auto_load("path/to/nested_directory")
        """
        for filename in os.listdir(path):
            filepath = os.path.join(path, filename)

            if os.path.isdir(filepath) and locale is Ellipsis:
                self.auto_load(filepath, locale=filename)

            if not os.path.isfile(filepath):
                continue

            lst = filename.rsplit(".", 1)
            if len(lst) < 2: continue
            name, suffix = lst

            if locale is None:
                if suffix == "json":    self.load_json(filepath, locale=name)
                elif suffix == "csv":   self.load_csv_i18n(filepath)
                elif suffix == "lang":  self.load_lang(filepath, locale=name)
                else: continue

            else:
                if suffix == "json":    self.load_json(filepath, locale=locale)
                elif suffix == "csv":   self.load_csv_i18n(filepath)
                elif suffix == "lang":  self.load_lang(filepath, locale=locale, superiors=name)
                else: continue



# Licensed under the MIT License.
# i18nco Copyright (C) 2022 numlinka.
# setup

# site
from setuptools import setup


setup(
    name = "i18nco",
    version = "1.5.0",
    description = "This is a simple and easy to use Python i18n library.",
    long_description = open("README_PyPI.md", "r", encoding="utf-8").read(),
    long_description_content_type = "text/markdown",
    author = "numlinka",
    author_email = "numlinka@163.com",
    url = "https://github.com/numlinka/pyi18nco",
    package_dir={"": "src"},
    packages = ["i18nco"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    license = "LGPLv3",
    keywords = ["sample", "i18n"],
    install_requires = []
)
